#define VL_CHECKING

#include "VLd.h"
#include "Solve.h"
#include "Quad2d.h"
#include "FuncBases.h"
#include "Tensor.h"
#include "BasisUtils.h"
#include <iomanip.h>
#include <stdio.h>
#include <string.h>


#define MTE_CHECK

#ifdef MTE_CHECK
#define PL_CHECK(P, L, n) CheckPushPull(P, L, n)
#define RS_CHECK(R, S, RULE) CheckTransfer(R, S, RULE)
#define FRS_CHECK(R, S) FCheckTransfer(R, S)
#else
#define PL_CHECK(P, L, n)
#define RS_CHECK(R, S, RULE)
#define FRS_CHECK(R, S)
#endif

// ------------------------------------------------------------

static Bool gVerbose = false;

/*
	Conversion between bases...

	Useful equations:
		
		[1] Matrix transfer Equation
		
		d = (Bdt Wd Bd)-1 Bdt Wd K Ws Bs s (MTE)
		d = Bd~ Bdt Wd K Ws Bs s 
				
		Bs is source basis: converts from source coefficients -> quadrature sample points
		Bd is destination basis: ditto. Bdt = trans(Bd).
		Ws is source weight matrix: contains quadrature coefficients along its
		diagonal.
		Wd is dest weight matrix: contains quadrature coefficients along its
		diagonal.
		
		[2] Basis conversions
		
		d =  (B2t W B2)-1 B2t W B1 s      (Conv from B1 -> B2)
		  =  B2-1 B1 s when B2 has an inv.
		W = weight matrix used for the common basis matrix.
		
		[3] 
		
		
	[Notes]
		
		(set of basis functions) x (quadrature method) -> Basis matrix.
		(quadrature method) -> Weight matrix
	
	Push:
		conversion between basis functions, essentially.
		but we split it up, because of piece-wise integral involved.
		
	Wavelet quadrature: essentially piece-wise quadrature...
	4 elt-> 16 elt: 16 x 4 conversion matrix.
	can: 
		(1) split quad into parts. Have 4 different samplings
	of B1 at the sub-quad quadrature points.
		(2) generate wavelet quadrature altogether. Same dif, really,
		but need massive weight matrix.
*/

// --- Routines for calculating push/pull matrices from bases & rules ------------------------------------

void MakePushPullMatrices(FunctionBasis &basis, QuadRule rule, Matd *P, Matd *L)
//  Works for tree wavelets, where the children of the parent set of basis functions
//	are orthogonal to each other.
{
	Matd		B, W, SB, T, U, A;
	Int			i, n = 1 << basis.FuncDim();	// n = 2 ^ d
	
	MakeBasisMatrix(basis, rule, B);
	MakeWeightMatrix(rule, W);
	T = trans(B) * W; 
	U = T * B;
	U = clamped(inv(U));

	for (i = 0; i < n; i++)
	{
		MakeSubBasisMatrix(basis, rule, i, 2, SB);	// binary/quad/octtree.
		// Set push matrix
		A = T * SB;
		P[i] = clamped(U * A);
		// Set pull matrix
		L[i] = clamped(U * trans(A)) / n;
	}
}

// Flatlet basis creation routines...

void LineConvertToFlatlet(FunctionBasis &basis, const QuadRule &flatRule, Int order, Matd P[2], Matd L[2])
{
	Matd	F, G;
	Int		i;
	
	MakeFlatletMatrices(basis, flatRule, HAAR_LINE, order, F, G);	
		
	for (i = 0; i < 2; i++)
	{
		P[i] = clamped(F * P[i] * G);
		L[i] = clamped(F * L[i] * G);
	}
}

void TriConvertToFlatlet(FunctionBasis &basis, const QuadRule &flatRule, Int order, Matd P[4], Matd L[4])
{
	Matd	F, G;
	Int		i;
	
	MakeFlatletMatrices(basis, flatRule, HAAR_TRI, order, F, G);	

	for (i = 0; i < 4; i++)
	{
		P[i] = clamped(F * P[i] * G);
		L[i] = clamped(F * L[i] * G);
	}
}

void FindMTEMatrices(
		FunctionBasis	&b,		// patch function basis for source & dest
		QuadRule		rule,	// quadrature rule to use for integration
		Matd			&R,		// output R + S
		Matd			&S
	)
// Finds R and S such that  T = R * K * S
// where K is the kernel sample matrix, and T the transport matrix.
{
	Matd		B, W, U;
	Int			i;

//	d = (Bdt Wd Bd)-1 Bdt Wd K Ws Bs s (MTE)
		
	MakeBasisMatrix(b, rule, B);
	MakeWeightMatrix(rule, W);

	U = clamped(inv(trans(B) * W * B));
	S = clamped(W * B);
	R = clamped(U * trans(B) * W);
}

void FindErrorMatrix(FunctionBasis &basis, QuadRule rule, Matd &R)
{
	// 
	
	Matd		B1, W, D1, I;
	Int			i;
		
	MakeBasisMatrix(basis, rule, B1);
	MakeWeightMatrix(rule, W);
	D1 = clamped(inv(trans(B1) * W * B1));
	
	R = clamped(B1 * D1 * trans(B1) * W);
	I.SetSize(R.Rows(), R.Cols());
	I = vl_I;
	
	// ... subtract off K(x,y)
}

#pragma mark -
// Routine for checking that a push followed by a pull is the identity function

void CheckPushPull(Matd *P, Matd *L, Int numChildren = 2)
// numChildren is the number of children each subdivision has.
// For binary/quadtree/octtree-type constructions, it'll be 2/4/8 respectively.
// This is tree-wavelet specific.
{
	if (!gVerbose) return;
	Int			i;
	Matd		result(L[0].Rows(), P[0].Cols());
	
	result = vl_zero;
	cout << ">>> push/pull is:" << endl;

	for (i = 0; i < numChildren; i++)
		result += clamped(L[i] * P[i], 1e-8);
		
	cout << clamped(result, 1e-6) << endl;
}

void CheckTransfer(Matd &R, Matd &S, const QuadRule &rule)
// Check that R K S, where K is the identity transfer matrix (Ws-1),
// is the identity transform.
{
	if (!gVerbose) return;
	Int			i;
	Matd		W;
	
	MakeWeightMatrix(rule, W);
	cout << ">>> transfer is:" << endl;
	cout << clamped(R * inv(W) * S, 1e-6) << endl;
}

void FCheckTransfer(Matd &R, Matd &S)
// Check that R K S, where K is the identity matrix,
// is the identity transform. (Flatlet transfer).
{
	if (!gVerbose) return;
	Int			i;

	cout << ">>> ftransfer is:" << endl;
	cout << clamped(R * S, 1e-6) << endl;
}

void CheckBasis(char *label, FunctionBasis &basis, QuadRule rule)
{
	if (!gVerbose) return;
	Matd B, W, D;
	MakeWeightMatrix(rule, W);
	MakeBasisMatrix(basis, rule, B);
	D = trans(B) * W * B;
	cout << ">>> " << label << ": I = " << clamped(inv(D) * D) << endl;
}



#pragma mark -
// --- Printing routines ------------------------------

void PrintVLMat(char *name, const Matd &M)
{
	Int i, j;
	Char *vlt;
	Bool vec;
	
	vec = (M.Cols() == 1);
	if (vec)
		vlt = "Vec";
	else
		vlt = "Mat";
	cout << "const ";

	if (M.Rows() <= 4 && (vec || M.Rows() == M.Cols()))
	{
		if (M.Rows() == 1)
		{
			cout << "Double " << name << " " << M[0][0] << endl;
			return;
		}
		else if (M.Rows() == 2)
			cout << vlt << "2d " << name << "(";
		else if (M.Rows() == 3)
			cout << vlt << "3d " << name << "(";
		else
			cout << vlt << "4d " << name << "(";
	}
	else
	{
		cout << vlt << "d " << name << "(" << M.Rows();
		if (!vec)
			cout << ", " << M.Cols();
		cout << ", ";
	}
		
	if (!vec)
		cout << endl;

	for (i = 0; i < M.Rows(); i++)
	{	
		if (!vec)
			cout << "        ";
		
		for (j = 0; j < M.Cols(); j++)
		{
			Double a = M[i][j];
			
			if (abs(Double(Int(a + 0.5)) - a) < 1e-7)
				cout << Int(a + 0.5) << ".0";
			else
				cout << a;
			if (j < M.Cols() - 1 || i < M.Rows() - 1)
				cout << ", ";
		}
		
		if (!vec)
			cout << endl;
	}
	
	if (vec)
		cout << ");" << endl << endl;
	else
		cout << "    );" << endl << endl;
}

void PrintNiceMat(char *name, const Matd &M)
{
	Int i, j;
		
	for (i = 0; i < M.Rows(); i++)
	{	
		if (i == 0)
		{
			cout << name << " = ";
			for (j = strlen(name); j < 5; j++)
				cout << ' ';
		}
		else
			cout << "        ";
		cout << "[";
		for (j = 0; j < M.Cols(); j++)
		{
			Double a = M[i][j];
			
			if (abs(Double(Int(a + 0.5)) - a) < 1e-7)
				cout << Int(a + 0.5) << ".0";
			else
				cout << a;
			if (j < M.Cols() - 1)
				cout << ", ";
		}
		
		cout << "]" << endl;
	}
}

void PrintPLMatrices(char *name, Matd *P, Matd *L, Int numChildren = 2)
{
	char temp[256];
	Int i;
	
	for (i = 0; i < numChildren; i++)	
	{
		sprintf(temp, "%s_P%d", name, i);
		PrintNiceMat(temp, P[i]);
	}
	for (i = 0; i < numChildren; i++)	
	{
		sprintf(temp, "%s_L%d", name, i);
		PrintNiceMat(temp, L[i]);
	}

	cout << endl;
}

void PrintPLVMatrices(char *name, Matd *P, Matd *L, Int numChildren = 2)
{
	char temp[256];
	Int i;
	
	for (i = 0; i < numChildren; i++)	
	{
		sprintf(temp, "%s_P%d", name, i);
		PrintVLMat(temp, P[i]);
	}
	for (i = 0; i < numChildren; i++)	
	{
		sprintf(temp, "%s_L%d", name, i);
		PrintVLMat(temp, L[i]);
	}
	
	cout << endl;
}

void PrintMTEMatrices(char *name, const Matd &R, const Matd &S)
{
	char temp[256];
	
	sprintf(temp, "%s_R", name);
	PrintVLMat(temp, R);
	sprintf(temp, "%s_S", name);
	PrintVLMat(temp, S);
	cout << endl;
}

void PrintErrorMatrix()
{
	Matd R;
	
	FindErrorMatrix(M2_QUAD, GL_QUAD_3, R);
	PrintVLMat("errorQ2", R);
}

Void RightTensorMult(const Matd &a, const Matd &c, Matd &d)
// d = a * C, where C is the tensor-product of matrix c with itself.
// O(mn^3).
{
	Int		n = c.Rows();
	Int		i, j, k, l, s, t, m;
	Matd	p;
	
	d.SetSize(a.Rows(), sqr(n));
	d = vl_0;
	p.SetSize(c);
	
	for (m = 0; m < a.Rows(); m++)
	{
		p = vl_0;
		for (l = 0; l < n; l++)
			for (i = 0, s = 0; i < n; i++)
				for (j = 0; j < n; j++, s++)
					p[i][l] += c[j][l] * a[m][s];
					
		for (k = 0, t = 0; k < n; k++)
			for (l = 0; l < n; l++, t++)
				for (i = 0; i < n; i++)
					d[m][t] += c[i][k] * p[i][l];
	}
}

#pragma mark -
// Header printing routines.

static Char *gHeaderMsg[] =
{
	"//",
	"//	File:		WaveCoeffs.cc   (auto-generated by MTE.cc)",
	"//",
	"//	Purpose:	Contains coefficients useful for wavelet/galerkin radiosity",
	"//",
	"//	Author:     Andrew Willmott, (c) 1996-1997",
	"//"
	"",
	"",
0 };


void PrintWaveletHeader()
{
	Matd		P[16], L[16];
	Matd		R, S;
	Int			i;
	Char		**curStr;
	Matd		M2_R, M3_R, M2_Tri_R, M3_Tri_R;
#ifdef MTE_S0
	cout << "// +--------------------------------+" << endl;
	cout << "// + Wavelet Push and Pull Matrices +" << endl;
	cout << "// +--------------------------------+" << endl;
	cout << endl;	
#else
	curStr = gHeaderMsg;
	while (*curStr)
		cout << *curStr++ << endl;
#endif

	CheckBasis("M2_LINE", M2_LINE, GL_2);
	CheckBasis("M3_LINE", M3_LINE, GL_3);
	CheckBasis("M4_LINE", M4_LINE, GL_4);

	CheckBasis("M2_TRI",  M2_TRI,  ST_TRI_2);
	CheckBasis("M2F_TRI", M2F_TRI, ST_TRI_3);
	CheckBasis("M3_TRI",  M3_TRI,  ST_TRI_5);
	CheckBasis("M3F_TRI", M3F_TRI, GA_TRI_7);
	CheckBasis("M4_TRI",  M4_TRI,  GA_TRI_7);

	cout << "// === Tensor Push/Pull matrices =========================================\n" << endl;

	MakePushPullMatrices(M2_LINE, GL_2, P, L);
	PrintPLVMatrices("M2", P, L);
	PL_CHECK(P, L, 2);
	
	LineConvertToFlatlet(M2_LINE, GL_2, 2, P, L);
	PrintPLVMatrices("F2", P, L);
	PL_CHECK(P, L, 2);

	MakePushPullMatrices(M3_LINE, GL_3, P, L);
	PrintPLVMatrices("M3", P, L);
	PL_CHECK(P, L, 2);

	LineConvertToFlatlet(M3_LINE, GL_3, 3, P, L);
	PrintPLVMatrices("F3", P, L);
	PL_CHECK(P, L, 2);

#ifdef MTE_LEV_4
	MakePushPullMatrices(M4_LINE, GL_4, P, L);
	PrintPLVMatrices("M4", P, L);
	PL_CHECK(P, L, 2);

	LineConvertToFlatlet(M4_LINE, GL_4, 4, P, L);
	PrintPLVMatrices("F4", P, L);
	PL_CHECK(P, L, 2);
#endif

	cout << "// === Tensor transfer matrices =========================================\n" << endl;

	FindMTEMatrices(M2_LINE, GL_2, R, S);
	PrintMTEMatrices("M2", R, S);
	RS_CHECK(R, S, GL_2);
	M2_R = R;
	
	FindMTEMatrices(M3_LINE, GL_3, R, S);	
	PrintMTEMatrices("M3", R, S);
	RS_CHECK(R, S, GL_3);
	M3_R = R;
	
	MakeFlatletMatrices(M2_LINE, GL_2, HAAR_LINE, 2, S, R);	
	PrintMTEMatrices("MCC2", R, S);
	FRS_CHECK(R, S);
	MakeFlatletMatrices(M3_LINE, GL_3, HAAR_LINE, 3, S, R);	
	PrintMTEMatrices("MCC3", R, S);
	FRS_CHECK(R, S);

	PrintVLMat("M2_x", sub(GL_QUAD_2, 0, 1, GL_QUAD_2.Rows(), 1));
	PrintVLMat("M2_y", sub(GL_QUAD_2, 0, 2, GL_QUAD_2.Rows(), 1));
	PrintVLMat("M3_x", sub(GL_QUAD_3, 0, 1, GL_QUAD_3.Rows(), 1));
	PrintVLMat("M3_y", sub(GL_QUAD_3, 0, 2, GL_QUAD_3.Rows(), 1));

	cout << "\n// === Triangular bases =========================================\n" << endl;
	
	MakePushPullMatrices(M2_TRI, ST_TRI_2, P, L);
	PrintPLVMatrices("M2_Tri", P, L, 4);
	PL_CHECK(P, L, 4);

	MakePushPullMatrices(M2F_TRI, ST_TRI_5, P, L);
	PL_CHECK(P, L, 4);
	TriConvertToFlatlet(M2F_TRI, ST_TRI_5, 2, P, L);
	PrintPLVMatrices("F2_Tri", P, L, 4);
	PL_CHECK(P, L, 4);

	MakePushPullMatrices(M3_TRI, ST_TRI_5, P, L);
	PrintPLVMatrices("M3_Tri", P, L, 4);
	PL_CHECK(P, L, 4);
	
	MakePushPullMatrices(M3F_TRI, GA_TRI_7, P, L);
	PrintPLVMatrices("M3F_Tri", P, L, 4);
	PL_CHECK(P, L, 4);
	TriConvertToFlatlet(M3F_TRI, GA_TRI_7, 3, P, L);

	PrintPLVMatrices("F3_Tri", P, L, 4);
	PL_CHECK(P, L, 4);

	cout << "// === Triangular transfer matrices =========================================\n" << endl;

	FindMTEMatrices(M2_TRI, ST_TRI_2, R, S);
	PrintMTEMatrices("M2_Tri", R, S);
	RS_CHECK(R, S, ST_TRI_2);
	M2_Tri_R = R;
	
	FindMTEMatrices(M3_TRI, ST_TRI_5, R, S);
	PrintMTEMatrices("M3_Tri", R, S);
	RS_CHECK(R, S, ST_TRI_5);
	M3_Tri_R = R;

	MakeFlatletMatrices(M2_TRI, ST_TRI_3, HAAR_TRI, 2, S, R);	
	PrintMTEMatrices("MCC2_Tri", R, S);
	FRS_CHECK(R, S);
	
	MakeFlatletMatrices(M3_TRI, ST_TRI_5, HAAR_TRI, 3, S, R);
	PrintMTEMatrices("MCC3_Tri", R, S);
	FRS_CHECK(R, S);
	
	PrintVLMat("MT2_x", sub(ST_TRI_2, 0, 1, ST_TRI_2.Rows(), 1)); // for M2
	PrintVLMat("MT2_y", sub(ST_TRI_2, 0, 2, ST_TRI_2.Rows(), 1));
	PrintVLMat("MT3_x", sub(ST_TRI_5, 0, 1, ST_TRI_5.Rows(), 1)); // for M3
	PrintVLMat("MT3_y", sub(ST_TRI_5, 0, 2, ST_TRI_5.Rows(), 1));
	
#if 1
	int j;
	
	Matd M2_EP(4, 2, 
		 1.0,  1.0,
		-1.0,  1.0,
		 1.0, -1.0,
		-1.0, -1.0,
		);
	Matd M2_E(4, 4), AA;
	
	PrintVLMat("M2_Ex", sub(M2_EP, 0, 0, M2_EP.Rows(), 1));
	PrintVLMat("M2_Ey", sub(M2_EP, 0, 1, M2_EP.Rows(), 1));

	for (i = 0; i < M2_QUAD.Dim(); i++)
		for (j = 0; j < M2_EP.Rows(); j++)
			M2_E[j][i] = M2_QUAD.Eval(i, M2_EP[j]);

	RightTensorMult(M2_E, M2_R, AA);
	PrintVLMat("M2_E", AA);
	
	Matd M3_E(4, 9);
	
	PrintVLMat("M3_Ex", sub(M2_EP, 0, 0, M2_EP.Rows(), 1));
	PrintVLMat("M3_Ey", sub(M2_EP, 0, 1, M2_EP.Rows(), 1));

	for (i = 0; i < M3_QUAD.Dim(); i++)
		for (j = 0; j < M2_EP.Rows(); j++)
			M3_E[j][i] = M3_QUAD.Eval(i, M2_EP[j]);

	RightTensorMult(M3_E, M3_R, AA);
	PrintVLMat("M3_E", AA);

// --- tri

	Matd M2T_EP(4, 2, 
		0.0, 1.0,
		0.33333, 0.33333,
		0.0, 0.0,
		1.0, 0.0,
		);
	Matd M2T_E(4, 3);
	
	PrintVLMat("M2T_Ex", sub(M2T_EP, 0, 0, M2T_EP.Rows(), 1));
	PrintVLMat("M2T_Ey", sub(M2T_EP, 0, 1, M2T_EP.Rows(), 1));

	for (i = 0; i < M2_TRI.Dim(); i++)
		for (j = 0; j < M2T_EP.Rows(); j++)
			M2T_E[j][i] = M2_TRI.Eval(i, M2T_EP[j]);

	PrintVLMat("M2T_E", M2T_E * M2_Tri_R);

	Matd M3T_E(4, 6);
	
	PrintVLMat("M3T_Ex", sub(M2T_EP, 0, 0, M2T_EP.Rows(), 1));
	PrintVLMat("M3T_Ey", sub(M2T_EP, 0, 1, M2T_EP.Rows(), 1));

	for (i = 0; i < M3_TRI.Dim(); i++)
		for (j = 0; j < M2T_EP.Rows(); j++)
			M3T_E[j][i] = M3_TRI.Eval(i, M2T_EP[j]);

	PrintVLMat("M3T_E", M3T_E * M3_Tri_R);

#endif
}

int main(int argc, char **argv)
{
	if (argc >= 2 && argv[1][0] == '-' && argv[1][1] == 'v')
		gVerbose = 1;
	PrintWaveletHeader();
	return(0);
}
